#include "Project.h"

#include "aimms/Include.h"

#include <sstream>
#include <assert.h>

#include "ErrorCollector.h"
#include "InternalException.h"
#include "CallBack.h"
#include "SliceAssigner.h"
#include "SliceRetriever.h"
#include "SliceDataCollector.h"

namespace aimms{


    Project::Project()
        : m_Session(0)
        , m_CallBack(0)
    {
    }
    Project::~Project()
    {
        if (m_CallBack){
            delete m_CallBack;
        }
    }

    const char* Project::getProjectName() const
    {
        return m_strProjectName.c_str();
    }

    int Project::open(const char* szAimmsLocation, const char* projectName, int timeOut, int maxFlushInBackground, int maxMemoryFlushInBackground, int maxBufferSize, ErrorCollector* ec)
    {
        m_strProjectName = projectName;

        IConfig* config = createDefaultConfig();

        try{
            config->setMaxFlush( maxFlushInBackground );
            config->setMaxBufferSize( maxBufferSize );
            config->setTimeout(timeOut);
            {
                std::ostringstream ossTmp;
                ossTmp << "{ \"InvocationTimeout\" : " << timeOut*1000 << " }";
                config->setCommunicationConfig(ossTmp.str().c_str());
            }
        } catch(std::exception& ex){
            ec->addError(ex.what());

        }
        // see if a logger-config-file exists:
        std::string strLoggerConfigFile = "aimmslink.logger.xml";
        FILE* fpFile = 0;
        errno_t err = fopen_s(&fpFile, strLoggerConfigFile.c_str(),"r");
        if (err == 0 && fpFile){
            fclose(fpFile);
            config->setLogConfigFileName(strLoggerConfigFile.c_str());
        }

        // see if a host logger-config-file exists:
        std::string strHostLoggerConfigFile = "host.logger.xml";
        fpFile = 0;
        err = fopen_s(&fpFile, strHostLoggerConfigFile.c_str(),"r");
        if (err == 0 && fpFile){
            fclose(fpFile);
            config->setHostLogConfigFileName( strHostLoggerConfigFile.c_str());     
        } 


        // createInstance may throw an exception, and since we cannot attach the callback until the session is created, 
        // we have to catch it separately.
        try{

            m_Session = openSession(szAimmsLocation , projectName, config);

        } catch (std::exception& ex){
            ec->addError(ex.what());
            delete config;
            return AM_General;
        } 
        delete config;
        m_CallBack = new CallBack(ec, projectName);
        m_Session->setCallBack(m_CallBack);

        return AM_OK;
    }

    int Project::assignSet(const char* name, int card, const char* const* labels, bool clearFirst)
    {
        try{
            ISetData* set = m_Session->openSet(name);
            assert(set);

            if (clearFirst){
                set->setLabels(labels, card);
            } else  {
                set->appendLabels(labels, card);
            }

        }catch(InternalException& ex){

            return ex.getCode();
        }
        return AM_OK;
    }

    template<typename T>
    int Project::retrieveScalar(IData* id, T* data)
    {
        IScalarData* scalar = dynamic_cast<IScalarData*>(id);
        if (!scalar) {
            m_CallBack->addError("This is not a scalar identifier");
            return AM_InvalidUse;
        }

        double value;
        scalar->getValue(value);
        *data = (float)value;
        scalar->close();
        return AM_OK;    
    }
    template<typename T>
    int Project::assignScalar(IData* id, T* data)
    {
        IScalarData* scalar = dynamic_cast<IScalarData*>(id);
        if (!scalar) {
            m_CallBack->addError("This is not a scalar identifier");
            return AM_InvalidUse;
        }

        double value = *data;
        scalar->setValue(value);
        scalar->close();
        return AM_OK;    
    }


    namespace{
        bool singleElementSlices(int** slices, int* sliceCards, int dim)
        {
            //returns true if each sliced level is sliced to only a single element
            for (int i = 0 ; i< dim ; ++i){
                if (slices[i]  && sliceCards[i]> 1) {
                    return false;  
                }
            }
            return true;
        } 
        void removeFullSlices(int dim, int** slices, int* slicecards, int* setcards)
        {
            // if a slice is in fact the whole set, remove it as a slice
            for(int i = 0; i< dim; ++i){
                if (slices[i] && (slicecards[i] == setcards[i])){
                    slices[i] = 0;
                }
            }
        }
        
        IDataView* createView(IMultiDimData* par, int** slices, int* cards, int& newcard)
        {
            //assumed are single element values as slices 
            //(only create a view if singleElementSlices returned true)
            newcard = 1;
            IFilter* filter =  par->createFilter();
            ITuple* elements = par->createTuple();
            int dim = par->getDimension();
            for(int i = 0 ; i< dim; ++i){
                if (slices[i]){
                    filter->restrict(i, slices[i][0]);
                } else {
                    newcard *= cards[i];
                }
            }
            IDataView* view = par->openView(filter);
            filter->close();
            elements->close();
            return view;
        }

    }
    bool Project::cardsOK(int* cards, IMultiDimData* par) {
        int dim = par->getDimension();
        for( int i = 0 ; i < dim; ++i){
            if (cards[i] != par->getIndexDomain(i)->getCardinality()){
                std::ostringstream oss;
                oss << "The cardinality of the indexdomain on level " << i << "(" << par->getIndexDomain(i)->getCardinality();
                oss << ") does not correspond with the supplied card (" << cards[i] << ")." ;
                m_CallBack->addError(oss.str().c_str());
                return false; 
            }
        }
        return true;
    }

    int Project::getIdentifier(const char* name, int dim, IData*& ret)
    {
        assert(m_Session);
        ret = m_Session->openData(name);
        if(!ret->getValueType() == vtDouble){
            m_CallBack->addError("This identifier is not double valued!");
            return AM_InvalidUse;
        }

        // the dimension of the identifier in aimms 
        // should match the amount of sets
        int actualdim = ret->getDimension();
        if (dim != actualdim) {
            std::ostringstream oss;
            oss << "The dimension of the identifier (" << actualdim << ") does not correspond with the supplied dimension (" << dim <<")." ;
            m_CallBack->addError(oss.str().c_str());
            return AM_InvalidUse;
        }
        return AM_OK;
    }

    template<typename T>
    int Project::assignIntIdentifier(const char* name, int dim, int* firsts, int* lasts, T* data, int** slices, int* slicecards)
    {
        IData* id=0; 
        try{
            int ret = getIdentifier(name,dim, id);
            if (ret > 0) {
                return ret;
            }
            if(id->getType() == tScalar){
                return assignScalar(id,data);
            }
            IMultiDimData* par = dynamic_cast<IMultiDimData*>(id);
            if (! par) {
                m_CallBack->addError("This is not a multidimensional identifier!");
                return AM_InvalidUse;
            }

            cacheIndexDomains(par);
            if(!slices){
                setValuesCMO(par, data, dim, firsts, lasts);
            } else{
                 int* setcards  = new int[dim];
                 for (int i = 0; i< dim; ++i){
	                setcards[i] =  lasts[i] - firsts[i] + 1 ;
                 }
                   
                removeFullSlices(dim, slices, slicecards, setcards);
               
                if( singleElementSlices(slices, slicecards, dim)){
                    
                    int newcard;
                    IDataView* view = createView(par, slices, setcards, newcard);
                   
                    SliceDataCollector<T> collector(par, data,newcard );
                    collector.SetIntegerLevelsCMO(slices, slicecards, firsts, lasts);
                    collector.collectData();
                    view->setValuesCMO(collector.getData(), newcard);
                    view->close();
                } else{
                    // the slice assigner assigns each individual value 
                    SliceAssigner<T> slicer(par, data);
                    slicer.SetIntegerLevelsCMO(slices, slicecards, firsts, lasts);
                    slicer.assignSlice();
                }
            }
            par->close();

            return AM_OK;

        } catch (InternalException& ex){
            if (id){
                id->close();
            }
            return ex.getCode();
        }
    }
    template<typename T>
    int Project::assignOrdIdentifier(const char* name, int dim, int* cards, T* data, int** slices, int* slicecards)
    {
        IData* id=0; 
        try{
            int ret = getIdentifier(name,dim, id);
            if (ret > 0) {
                return ret;
            }
            if(dim == 0){
                return assignScalar(id,data);
            }
            IMultiDimData* par = dynamic_cast<IMultiDimData*>(id);
            if (! par) {
                m_CallBack->addError("This is not a multidimensional identifier!");
                return AM_InvalidUse;
            }
            cacheIndexDomains(par);
            if (!cardsOK(cards,par)) {
                return AM_InvalidUse;
            }
            if(!slices){
                setValuesCMO(par, data, dim, cards);
            } else{ 

                removeFullSlices(dim, slices, slicecards, cards);
               
                if( singleElementSlices(slices, slicecards, dim)){
                    
                    int newcard;
                    IDataView* view = createView(par, slices, cards, newcard);
                   
                    SliceDataCollector<T> collector(par, data,newcard );
                    collector.SetOrdinalLevelsCMO(slices, slicecards, cards);
                    collector.collectData();
                    view->setValuesCMO(collector.getData(), newcard);
                    view->close();
                } else{
                    SliceAssigner<T> slicer(par, data);
                    slicer.SetOrdinalLevelsCMO(slices, slicecards, cards);
                    slicer.assignSlice();
                }
            }
            par->close();
            return AM_OK;

        } catch (InternalException& ex){
            if (id){
                id->close();
            }
            return ex.getCode();
        }

    }
    int Project::runProcedure(const char* name, int& returncode)
    {

        IProcedure* proc = 0;
        try{
            IProcedure* proc = m_Session->openProcedure(name);
            if (!proc){
                std::ostringstream oss;
                oss << "could not find procedure with name '"<< name << "'" ;
                m_CallBack->addError(oss.str().c_str());
                return AM_InvalidUse;
            }
            m_Session->closeAllData(); 
            m_SetCache.clear();

            returncode = proc->run();
            proc->close();
        } catch (InternalException& ex){
            if(proc){
                proc->close();
            }
            return ex.getCode();
        }
        return AM_OK;
    }
    template<typename T>
    int Project::retrieveIntIdentifier(const char* name, int dim, int* firsts, int* lasts, T* data, int** slices, int* slicecards)
    {
        int errorcode;
        IData* id;


        try{
            errorcode = getIdentifier(name,dim, id);
            if (errorcode > 0) {
                return errorcode;
            }
            if (dim == 0) 
                return retrieveScalar(id,data);

            IMultiDimData* par = dynamic_cast<IMultiDimData*>(id);
            if (! par) {
                m_CallBack->addError("This is not a multidimensional identifier!");
                return AM_InvalidUse;
            }
            cacheIndexDomains(par);
            SliceRetriever<T> slicer(par,data,500);
            slicer.SetIntegerLevelsCMO(slices, slicecards, firsts, lasts);

            if (slices != 0) {
                slicer.clearSlice();
            } else { 
                int card = 0;
                std::vector<int> setSize(dim);
                for (int lvl = 0; lvl< dim; ++lvl) {
                    setSize[lvl] = 1+lasts[lvl]-firsts[lvl];
                    card *= setSize[lvl];
                } 
                memset(data,0,card*sizeof(float));
            }
            slicer.retrieveSlice();
            // make sure the slicer is destroyed before par is closed, since it closes one of par's iterators

        } catch (InternalException& ex){
            if(id) {
                id->close();
            }
            return ex.getCode();
        }
        id->close();
        return AM_OK;
    }
    template<typename T>
    int Project::retrieveOrdIdentifier(const char* name, int dim, int* cards, T* data, int** slices, int* slicecards)
    {
        int errorcode;
        IData* id;

        errorcode = getIdentifier(name,dim, id);
        if (errorcode > 0) {
            return errorcode;
        }
        if (dim == 0) 
            return retrieveScalar(id,data);

        try{
            IMultiDimData* par = dynamic_cast<IMultiDimData*>(id);
            if (! par) {
                m_CallBack->addError("This is not a multidimensional identifier!");
                return AM_InvalidUse;
            }
            if (!cardsOK(cards,par)) {
                return AM_InvalidUse;
            }
            cacheIndexDomains(par);
            SliceRetriever<T> slicer(par,data,500);
            slicer.SetOrdinalLevelsCMO(slices, slicecards, cards);

            // full data; use next
            if (slices == 0) {
                int totalcard = 1;
                for( int d = 0 ; d < dim; ++d){
                    totalcard *= cards[d];
                }
                memset(data,0,totalcard*sizeof(float));
            } else {
                slicer.clearSlice();
            }
            slicer.retrieveSlice();
            // make sure the slicer is destroyed before par is closed, since it closes one of par's iterators

        } catch (InternalException& ex){
            if(id) {
                id->close();
            }
            return ex.getCode();
        }
        id->close();
        return AM_OK;
    }
    int Project::close()
    {
        int exitCode = AM_OK;

        if(m_Session){
            try{
                // flush all not yet committed data to AIMMS
                // NOTE: the close method should also does this internally,
                // unfortunately the close does not terminate properly when 
                // there is an error in the data that was to be flushed,
                // so as work around do this explicitly here
                m_Session->updateData();
            } catch (InternalException& ex){
                exitCode = ex.getCode();
            }

            try{
                m_Session->close();
            } catch (InternalException& ex){
                if (exitCode == AM_OK){
                    exitCode = ex.getCode();
                }
            }
        }

        delete this;
        return exitCode;
    }

    void Project::cacheIndexDomains(IDataView* par)
    {
        int dim = par->getDimension();

        for(int i = 0; i < dim; ++i){
            const ISetData* set = par->getIndexDomain(i);
            setOfISetP::iterator found = m_SetCache.find(set);
            if (found == m_SetCache.end()){
                m_SetCache.insert(m_Session->openSet(set->getName()));
            }

        }
    }
    template<typename T>
    void Project::setValuesCMO(IDataView* par,  T* data, int dim, int* firsts, int* lasts)
    {
        int* cards = new int[dim];
        for (int i = 0 ; i < dim; ++i){
            cards[i] = 1+  (lasts[i] - firsts[i]);
        }
        setValuesCMO(par, data, dim, cards);
        delete [] cards;
    }
    template<typename T>
    void Project::setValuesCMO(IDataView* par,  T* data, int dim, int* cards)
    {
        int card = 1;
        for (int i = 0 ; i < dim; ++i){
            card *= cards[i];
        }

        double* values = new double[card];
        for (int i = 0 ; i < card; ++i){
            values[i] = (double) data[i];
        }
        par->setValuesCMO(values, card);
        delete [] values;
    }
    

    template int Project::assignIntIdentifier<float>(const char* name, int dim, int* firsts, int* lasts, float* data, int** slices, int* slicecards);
    template int Project::assignIntIdentifier<double>(const char* name, int dim, int* firsts, int* lasts, double* data, int** slices, int* slicecards);
    template int Project::assignOrdIdentifier<float>(const char* name, int dim, int* cards, float* data, int** slices, int* slicecards);
    template int Project::assignOrdIdentifier<double>(const char* name, int dim, int* cards, double* data, int** slices, int* slicecards);
    template int Project::retrieveIntIdentifier<float>(const char* name, int dim, int* firsts, int* lasts, float* data, int** slices, int* slicecards);
    template int Project::retrieveIntIdentifier<double>(const char* name, int dim, int* firsts, int* lasts, double* data, int** slices, int* slicecards);
    template int Project::retrieveOrdIdentifier<float>(const char* name, int dim, int* cards, float* data, int** slices, int* slicecards);
    template int Project::retrieveOrdIdentifier<double>(const char* name, int dim, int* cards, double* data, int** slices, int* slicecards);

}