#ifndef _AIMMS_LINK_H
#define _AIMMS_LINK_H

#ifdef AIMMSLINK_EXPORTS
#define AIMMSLINK_API __declspec(dllexport)
#else
#define AIMMSLINK_API __declspec(dllimport)
#endif

// aimmsLink.h declares the external C interface for use with Fortran
extern "C"{
    /* all functions return 0 if no error occured, an error code 
    otherwise. If an error occured the 'AimmsGetLastErrorMessage' 
    can be used to retrieve an description of the specified error. */


    /* AimmsOpenProject establishes the connection with AIMMS, opens the 
    project and associates a handle with the project. */
    AIMMSLINK_API int AimmsOpenProject(
        const char* aimmsFolder, 
        const char* projectName, 
        int* projectHandle, 
        int timeOut,
        int maxFlushInBackground, 
        int maxMemoryFlushInBackground,  
        int maxBufferSize );

    /* AimmsCloseProject closes the project associated with the projectHandle
    and terminates the connection with AIMMS. */
    AIMMSLINK_API int AimmsCloseProject( 
        int projectHandle );

    /* AimmsRunProcedure runs the specified AIMMS procedure and receives the
    return code in the out-argument 'returnCode'. */
    AIMMSLINK_API int AimmsRunProcedure( 
        int projectHandle, 
        const char* procedureName, 
        int* returnCode );

    /* AimmsAssignSet adds the labels to the named set.
    In case the set is a subset and labels do not yet exist in its super
    set, they are added there as well. Labels already existing in the set
    are skipped. The boolean clearFirst should be set to true if 
    the set is to be emptied before adding the specified elements.
    An error is returned if the setName does not exist in the model, or 
    does not refer to a set. */
    AIMMSLINK_API int AimmsAssignSet(
        int projectHandle, 
        const char* setName, 
        int card, 
        const char* const* labels, 
        bool clearFirst);



    /* for all identifier functions, there is an Int and an Ord variant.
    Int variants: the positions are interpreted as integer values. 
    All indexdomains of the identifier must be subsets of 
    the rootset "Integers".
    Ord variants: the set positions are interpreted as ordinals of the
    indexdomains of the identifier.
    */                     


    /* Assign Identifier assigns the specified data to the named identifier.
    An error is returned if the AIMMS identifier is not real valued, or if 
    its dimension does not correspond to the number of sets.
    Assumed is a column major ordering. */
    AIMMSLINK_API int AimmsIntAssignIdentifier(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* firsts, 
        int* lasts, 
        float* identifier);

    AIMMSLINK_API int AimmsIntAssignIdentifierDouble(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* firsts, 
        int* lasts, 
        double* identifier);

    AIMMSLINK_API int AimmsOrdAssignIdentifier(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* cards,
        float* identifier);

    AIMMSLINK_API int AimmsOrdAssignIdentifierDouble(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* cards,
        double* identifier);

    /* Assign Identifier Slice assigns the data slice to the named identifier.
    An error is returned if the AIMMS identifier is not real valued, or if 
    its dimension does not correspond to the number of sets.
    Assumed is a column major ordering. */
    AIMMSLINK_API int AimmsIntAssignIdentifierSlice(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* firsts,
        int* lasts, 
        float* identifier, 
        int** slices=0, 
        int* sliceCardinalities=0);

    AIMMSLINK_API int AimmsIntAssignIdentifierSliceDouble(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* firsts,
        int* lasts, 
        double* identifier, 
        int** slices=0, 
        int* sliceCardinalities=0);

    AIMMSLINK_API int AimmsOrdAssignIdentifierSlice(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* cards,
        float* identifier, 
        int** slices=0, 
        int* sliceCardinalities=0);

    AIMMSLINK_API int AimmsOrdAssignIdentifierSliceDouble(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* cards,
        double* identifier, 
        int** slices=0, 
        int* sliceCardinalities=0);

    /* Retrieve Identifier retrieves the data of the specified AIMMS identifier
    An error is returned if the parameter is not double valued, and if the
    dimension of the AIMMS identifier does not correspond to the number of
    sets. Assumed is a column major ordering. */
    AIMMSLINK_API int AimmsIntRetrieveIdentifier(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* firsts, 
        int* lasts, 
        float* identifier);

    AIMMSLINK_API int AimmsIntRetrieveIdentifierDouble(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* firsts, 
        int* lasts, 
        double* identifier);

    AIMMSLINK_API int AimmsOrdRetrieveIdentifier(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* cards,
        float* identifier);

    AIMMSLINK_API int AimmsOrdRetrieveIdentifierDouble(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* cards,
        double* identifier);

    /* Retrieve Identifier Slice retrieves the data of a slice of the specified
    AIMMS identifier.
    An error is returned if the parameter is not double valued, and if the
    dimension of the AIMMS identifier does not correspond to the number of
    sets. 
    Assumed is a column major ordering. */
    AIMMSLINK_API int AimmsIntRetrieveIdentifierSlice(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* firsts, 
        int* lasts, 
        float* identifier, 
        int** slices=0,
        int* sliceCardinalities=0);

    AIMMSLINK_API int AimmsIntRetrieveIdentifierSliceDouble(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* firsts, 
        int* lasts, 
        double* identifier, 
        int** slices=0,
        int* sliceCardinalities=0);

    AIMMSLINK_API int AimmsOrdRetrieveIdentifierSlice(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* cards,
        float* identifier, 
        int** slices=0, 
        int* sliceCardinalities=0);

    AIMMSLINK_API int AimmsOrdRetrieveIdentifierSliceDouble(
        int projectHandle, 
        const char* identifierName, 
        int dimension, 
        int* cards,
        double* identifier, 
        int** slices=0, 
        int* sliceCardinalities=0);

    /* AimmsGetLastErrorMessage writes the last error message to the buffer if
    it is large enough. 'textBufferLen' is an in-out argument, which should 
    hold the length of the provided buffer, and receives the length of the
    last error message. */
    AIMMSLINK_API int AimmsGetLastErrorMessage(
        int* textBufferLen, 
        char* textBuffer );

}
#endif //end #ifndef _AIMMS_LINK_H